/*
 * ps2.c - an interface library for ps2 devices.  Common devices are
 * mice, keyboard, barcode scanners etc.  See the examples for mouse and
 * keyboard interfacing.
 * limitations:
 *		Works for me - will probably not work for you.
 *      we do not handle parity errors.
 *      The timing constants are hard coded from the spec, but most
 *		timings actually come from the connected device itself.
 *      lots of room for optimization.
 *		The ps2read routine has some really strange bugfix which requires
 *		String to Bitstring transcoding. Since transcoding takes time,
 *		one must pass the ps2read routine the number of expected packets.
 *		I found out this is due to pins NEVER get HIGH. So you must check for !=LOW instead.
 *		Note: If you pass more, the timer will run out, returning the Number Zero
 *		Note: If you pass less, everything will look OK, but if you call the
 *		routine again while the device is still sending and not run out the
 *		timer (which is kinda difficult), you will get thrown out of sync.
 *		(this is the main reason why there is a timer in the first place)
 */

#include "ps2.h"
 
#include <util/delay.h>

void pinMode(int pin, int mode)
{
	//pin--; //Counting starts with 1 :-(
	if (mode==INPUT)
	{
		DDRD &= (~(1<<pin)); //Pin in Register auf Input setzen (0=Input, default)
	}
	else
	{
		DDRD |= (1<<pin); //Pin in Register auf Output setzen (1=Output)
	}
}

void digitalWrite(int pin, int value) //seems to work OK
{
	//pin--;
	if(value==HIGH)
	{
		PORTD |= (1<<pin);
	}else
	{
		PORTD &= ~(1<<pin);
	}
}

int digitalRead(int pin) //works but exhibits strange behaviour sometimes (on startup, always reports H for a while, until you really FORCE the signal down for quite a while, after that, it worked just fine, but this happens rarely and is probably my fault.)
{
	//pin--;
	//ps2_gohi(pin); //sollte eigentlich hochohmig sein...
	return PIND & (1<<pin);
}

void ps2_gohi(int pin) //gohi and golo were tested working fine.
{
	digitalWrite(pin, HIGH);
	pinMode(pin, INPUT);
}

void ps2_golo(int pin)
{
	digitalWrite(pin, LOW);
	pinMode(pin, OUTPUT);//Erst Schreiben, DANN auf Output setzen, laut AVR Tutorial entsteht sonst ein ungewollter LOW Puls
}

uint32_t my_ps2read(int numpackets)
{
	//uint8_t result;
	uint32_t bit;
	uint32_t data;
	uint8_t devid;
	uint8_t st;
	//uint8_t x;
	
	uint16_t time;
	
	char Buffer[48];
	char temp;
	char i;
	
	temp=numpackets*11; //so the AVR does that heavy multiplication only once
	devid=0;
	time=0;

	//ps2_gohi(_ps2clk); //NOTE: if(!work) delete this line

	while(devid<temp&&time<16384)
	{
		if(time==0) PORTB &= 0xFB; else PORTB |= 4; //error on while waiting for transmission. If it stays on, you know there's something wrong..
		//TODO: If start bit==HIGH, you're too early! consume it and you'll get a valid string.
		if(digitalRead(_ps2clk)==LOW) //this works better than ps2_read, since it does NOT always return 0
		{ //NOTE: this sometimes (most of the time) misses the start bit. Everything gets shifted left. Also, it is not too stable, I always get 12 Bits, so with start bit, it's 13, so we know there's something wrong.
			if(digitalRead(_ps2data)==LOW) //OK TODO check for !=LOW instead, because it will never be HIGH for some reason.
			{
				//lcd_data('L'); //uhh... LCD is too slow, too.
				//uart_putc('L'); //fuck UART. WAAAY too slow.
				Buffer[devid]='L'; //OK this wastes 8 times the space, but it works.
			}
			else
			{
				//uart_putc('H');
				//lcd_data('H');
				Buffer[devid]='H';
				//data = data | bit; //<- THIS DOES NOT WORK HERE FOR SOME UNKNOWN REASON! (data will always be zero on my machine)
			}
			while (digitalRead(_ps2clk) == LOW); //wait for clk to go high again.
			//bit = bit << 1;
			devid++; //LCD line break.
			time=0; //so kann ich irgendwann stoppen, wieviel Zeit vom ->Ende<- einer Transmission zum Anfang der nchsten vergeht.
		}
		if(devid>0) time++; //ok zeit luft erst, wenn Beginn der Transmission.
	}
	Buffer[devid]=0; //\0 to terminate string.. +1 is one too far, so we get up to one junk byte.
	//uart_puts(Buffer); LLHLHLHLHH
	bit=1;
	data=0;
	//if(time>16384) lcd_data('T');
	//assume 00101010111 00101010111 00101010111 00101010111
	//2 10 13 21 24 32 35 43
	//2 13 24 35 == 2+11*(numpackets-1)
	//10 21 32 43 = 10+11*(numpackets-1)
	
	//uart_puts(Buffer);
	
	for(i=0; i<numpackets; i++) //0..3
	{
		char a=2+11*i;
		char e=10+11*i;
		if(devid<e) {return 0;}
		for(st=devid-e;  st<devid-a; st++) //extract the 8 data bits
		{
			if(Buffer[st]=='H')
			{
				data |= bit; 
			}
			bit<<=1;
			//uart_putc(Buffer[st]);
		}
	}
//	uart_putc(':');
//	itoa(data, Buffer, 16);
//	uart_puts(Buffer);
//	uart_putc(13);
//	uart_putc(10); //btw. the order is wurscht.
	devid=0;
	return data;
}

void my_ps2write(unsigned char data)
{
	/*
    1)   Bring the Clock line low for at least 100 microseconds.
    2)   Bring the Data line low.
    3)   Release the Clock line.
    4)   Wait for the device to bring the Clock line low.
    5)   Set/reset the Data line to send the first data bit
    6)   Wait for the device to bring Clock high.
    7)   Wait for the device to bring Clock low.
    8)   Repeat steps 5-7 for the other seven data bits and the parity bit
    9)   Release the Data line.
    10) Wait for the device to bring Data low.
    11) Wait for the device to bring Clock  low.
    12) Wait for the device to release Data and Clock
	*/
	unsigned char i;
	unsigned char parity = 1;
	i=0;
	ps2_golo(_ps2clk);
	_delay_us(200);
	ps2_golo(_ps2data);
	ps2_gohi(_ps2clk);
	//when clk is read low, send bit. mouse will grab bit while clk=high
	while(i<8)
	{
		if(i==0) PORTB &= 0xFB; else PORTB |= 4; //turn on error while waiting for device to send clock. As said: If it stays on, there's something wrong.
		if(digitalRead(_ps2clk)==LOW)
		{
			if (data & 0x01) //let's just hope that works and I don't have to do that 'H'&'L' Bullfuck again
			{
				ps2_gohi(_ps2data);
			} else {
				ps2_golo(_ps2data);
			}			
			i++;
			parity = parity ^ (data & 0x01);
			data = data >> 1;
			while (digitalRead(_ps2clk) == LOW);
		}
	}
	i=0;
	while(i==0)
	{
		if(digitalRead(_ps2clk)==LOW)
		{
			if (parity)
			{
				ps2_gohi(_ps2data);
			} else {
				ps2_golo(_ps2data);
			}
			while (digitalRead(_ps2clk) == LOW);
			i++;
		}	
	}
	//yeah I know, debugging this is worst. hm... why don't I put that code on another micro and use this one for receiving?
	ps2_gohi(_ps2data);
	//_delay_us(5); //dangerous, since mouse IMMEDIATELY sends response (holding clk low probably resets device)
	i=0;
	while(i<2)
	{
		if(digitalRead(_ps2clk)==LOW)
		{
			while (digitalRead(_ps2clk) == LOW); //eat stop and ack bit.
			i++;
		}	
	}
	
	//uart_puts("done.");
	//Do I really have to read that fucking ACK!?
	//ps2_golo(_ps2clk); //NOTE: if(!work) delete this line

}

/* write a byte to the PS2 device */
/*
void ps2_write(unsigned char data)
{
	unsigned char i;
	unsigned char parity = 1;
	
	ps2_gohi(_ps2data);
	ps2_gohi(_ps2clk);
	_delay_us(200); //originally 300, manual says at least 100.
	ps2_golo(_ps2clk);
	_delay_us(200);
	ps2_golo(_ps2data);
	_delay_us(10);
	ps2_gohi(_ps2clk);	// start bit
	// wait for device to take control of clock
	while (digitalRead(_ps2clk) == HIGH)
		;	// this loop intentionally left blank
	// clear to send data
	for (i=0; i < 8; i++)
	{
		if (data & 0x01)
		{
			ps2_gohi(_ps2data);
		} else {
			ps2_golo(_ps2data);
		}
		// wait for clock
		while (digitalRead(_ps2clk) == LOW)
			;
		while (digitalRead(_ps2clk) == HIGH)
			;
		parity = parity ^ (data & 0x01);
		data = data >> 1;
	} //btw. o'scope tells me, this works (when no device is connected, data is sent and looks sane.). But it shouldn't since there is no Clock to wait for - it should become an endless loop.
	// parity bit
	if (parity)
	{
		ps2_gohi(_ps2data);
	} else {
		ps2_golo(_ps2data);
	}
	// clock cycle - like ack.
	while (digitalRead(_ps2clk) == LOW)
		;
	while (digitalRead(_ps2clk) == HIGH)
		;
	// stop bit
	ps2_gohi(_ps2data);
	_delay_us(50);
	while (digitalRead(_ps2clk) == HIGH)
		;
	// mode switch
	while ((digitalRead(_ps2clk) == LOW) || (digitalRead(_ps2data) == LOW))
		;
	// hold up incoming data
	ps2_golo(_ps2clk);
}
*/

/*
 * read a byte of data from the ps2 device.  Ignores parity.
 */
 /*
unsigned char ps2_read(void)
{
	unsigned char data;
	unsigned char i;
	unsigned char bit;

	data=0;
	bit=1;
	
	// start clock
	ps2_gohi(_ps2clk);
	ps2_gohi(_ps2data);
	_delay_us(50);
	while (digitalRead(_ps2clk) == HIGH)
		;
	//_delay_us(5);	// not sure why. what? wait for edge to fall/rise! Can an AVR@8 MHz wait that short?
	//should put some NOPs here.



	
	while (digitalRead(_ps2clk) == LOW)
		;	// eat start bit
	for (i=0; i < 8; i++)
	{
		while (digitalRead(_ps2clk) == HIGH)
			;
			
		if (digitalRead(_ps2data) == HIGH)
		{
			data |= bit;
		}
		while (digitalRead(_ps2clk) == LOW)
			;
		bit <<=1;
	}
	// eat parity bit, ignore it.
	while (digitalRead(_ps2clk) == HIGH)
		;
	while (digitalRead(_ps2clk) == LOW)
		;
	// eat stop bit
	while (digitalRead(_ps2clk) == HIGH)
		;
	while (digitalRead(_ps2clk) == LOW)
		;
	ps2_golo(_ps2clk);	// hold incoming data - ah there is the reason, clk is held lo after transmit. I was wondering...

	return data;
}

*/